const stringifyPath = (path, context) => {
    var _a;
    return path.reduce((acc, segment, index) => {
        var _a;
        if (index < ((_a = context === null || context === void 0 ? void 0 : context.cutOff) !== null && _a !== void 0 ? _a : 0))
            return acc;
        if (typeof segment === 'string')
            return `${acc}.${segment}`;
        return `${acc}[${segment}]`;
    }, `${' '.repeat(((_a = context === null || context === void 0 ? void 0 : context.depth) !== null && _a !== void 0 ? _a : 0) * 2)}$`);
};
export function isString(exact) {
    const error = (path) => ({
        expected: 'string',
        path,
        toString: (context) => {
            return `${stringifyPath(path, context)}: expected string`;
        }
    });
    if (exact !== undefined)
        return (value, path) => {
            if (value !== exact)
                throw error(path !== null && path !== void 0 ? path : []);
            return true;
        };
    return (value, path) => {
        if (typeof value !== 'string')
            throw error(path !== null && path !== void 0 ? path : []);
        return true;
    };
}
export function isNumber(exact) {
    const error = (path) => ({
        expected: 'number',
        path,
        toString: (context) => `${stringifyPath(path, context)}: expected number`,
    });
    if (exact !== undefined)
        return (value, path) => {
            if (value !== exact)
                throw error(path !== null && path !== void 0 ? path : []);
            return true;
        };
    return (value, path) => {
        if (typeof value !== 'number')
            throw error(path !== null && path !== void 0 ? path : []);
        return true;
    };
}
export function isBoolean(exact) {
    const error = (path) => ({
        expected: 'boolean',
        path,
        toString: (context) => `${stringifyPath(path, context)}: expected boolean`,
    });
    if (exact !== undefined)
        return (value, path) => {
            if (value !== exact)
                throw error(path !== null && path !== void 0 ? path : []);
            return true;
        };
    return (value, path) => {
        if (typeof value !== 'boolean')
            throw error(path !== null && path !== void 0 ? path : []);
        return true;
    };
}
export const isNull = (value, path) => {
    if (value !== null)
        throw {
            expected: 'null',
            path,
            toString: (context) => `${stringifyPath(path !== null && path !== void 0 ? path : [], context)}: expected null`,
        };
    return true;
};
export const isUndefined = (value, path) => {
    if (value !== undefined)
        throw {
            expected: 'undefined',
            path,
            toString: (context) => `${stringifyPath(path !== null && path !== void 0 ? path : [], context)}: expected undefined`,
        };
    return true;
};
export function isArray(schema) {
    const error = (path) => ({
        expected: 'array',
        path,
        toString: (context) => `${stringifyPath(path, context)}: expected array`,
    });
    if (schema === undefined)
        return (value, path) => {
            if (!Array.isArray(value))
                throw error(path !== null && path !== void 0 ? path : []);
            return true;
        };
    return (value, path) => {
        if (!Array.isArray(value))
            throw error(path !== null && path !== void 0 ? path : []);
        value.forEach((element, index) => schema(element, [...path !== null && path !== void 0 ? path : [], index]));
        return true;
    };
}
export function has(key, schema) {
    return (value, path) => {
        if (typeof value !== 'object' || Array.isArray(value))
            throw {
                expected: 'object',
                path,
                toString: (context) => `${stringifyPath(path !== null && path !== void 0 ? path : [], context)}: expected object`,
            };
        return schema(value[key], [...path !== null && path !== void 0 ? path : [], key]);
    };
}
export function at(key, schema) {
    return (value, path) => {
        if (!Array.isArray(value))
            throw {
                expected: 'array',
                path,
                toString: (context) => `${stringifyPath(path !== null && path !== void 0 ? path : [], context)}: expected array`,
            };
        return schema(value[key], [...path !== null && path !== void 0 ? path : [], key]);
    };
}
export function oneOf(...args) {
    return (value, path) => {
        const errors = [];
        for (const schema of args)
            try {
                schema(value, path !== null && path !== void 0 ? path : []);
            }
            catch (error) {
                errors.push(error);
            }
        if (errors.length < args.length)
            return true;
        throw {
            expected: 'one-of',
            errors,
            path,
            toString: (context) => {
                var _a, _b, _c;
                const innerContext = {
                    depth: ((_a = context === null || context === void 0 ? void 0 : context.depth) !== null && _a !== void 0 ? _a : 0) + 1,
                    cutOff: (_b = path === null || path === void 0 ? void 0 : path.length) !== null && _b !== void 0 ? _b : 0,
                };
                return `${stringifyPath(path !== null && path !== void 0 ? path : [], context)}: one-of [\n${errors.reduce((acc, error) => {
                    if (acc === undefined)
                        return error.toString(innerContext);
                    return `${acc}\n${error.toString(innerContext)}`;
                }, undefined)}\n${' '.repeat(((_c = context === null || context === void 0 ? void 0 : context.depth) !== null && _c !== void 0 ? _c : 0) * 2)}]`;
            }
        };
    };
}
export function allOf(...args) {
    return (value, path) => {
        const errors = [];
        for (const schema of args)
            try {
                schema(value, path !== null && path !== void 0 ? path : []);
            }
            catch (error) {
                errors.push(error);
            }
        if (errors.length === 0)
            return true;
        throw {
            expected: 'all-of',
            errors,
            path,
            toString: (context) => {
                var _a, _b, _c;
                const innerContext = {
                    depth: ((_a = context === null || context === void 0 ? void 0 : context.depth) !== null && _a !== void 0 ? _a : 0) + 1,
                    cutOff: (_b = path === null || path === void 0 ? void 0 : path.length) !== null && _b !== void 0 ? _b : 0,
                };
                return `${stringifyPath(path !== null && path !== void 0 ? path : [], context)}: all-of [\n${errors.reduce((acc, error) => {
                    if (acc === undefined)
                        return error.toString(innerContext);
                    return `${acc}\n${error.toString(innerContext)}`;
                }, undefined)}\n${' '.repeat(((_c = context === null || context === void 0 ? void 0 : context.depth) !== null && _c !== void 0 ? _c : 0) * 2)}]`;
            }
        };
    };
}
